;; Copyright (C) 2024  Felisp
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Affero General Public License as published by
;; the Free Software Foundation, version 3 of the License.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Affero General Public License for more details.
;;
;; You should have received a copy of the GNU Affero General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

(ns rss-thread-watch.core
  (:require [clojure.java.io :as io]
            [clojure.edn :as edn]
            [clojure.tools.cli :refer [parse-opts]]
            [ring.adapter.jetty :as jetty]
            [ring.middleware.params :as rp]
            [rss-thread-watch.watcher :as watcher]
            [rss-thread-watch.feed-generator :as feed]
            [rss-thread-watch.utils :as u])
  (:gen-class))

(def VERSION "0.4.0")

;; Internal default config
(def CONFIG-DEFAULT
  "Internal default config"
  {:port 6969
   :default-board "/mlp/"
   :enable-board-listing true
   :board-disabled-message "This board is not enabled for feed generation.\n\nYou can contact me here: [contact] and I may enable it for you"
   :boards-defaults {:refresh-rate 300
                     :starting-page 7
                     :default-chod 94
                     :target "https://api.4chan.org"
                     :lazy-load true}
   :boards-enabled {"/mlp/" {:lazy-load false}
                    "/g/" {:starting-page 7}
                    "/po/" {:starting-page 8
                            :refresh-rate 86400}
                    "/p/" {:starting-page 8
                           :refresh-rate 1800}}})

(def cli-options
  "Configuration defining program arguments for cli.tools"
  [["-v" "--version" "Print version and license information"]
   ["-h" "--help" "Prints help"]
   ["-c" "--config CONFIG_FILE" "Specify config file to use for this run"
    :default "./config.edn"
    :validate [#(u/file-exists? %) "Specified config file does not exist or is not readable"]]
   [nil "--print-default-config" "Prints internal default config file to STDOUT and exits"]])

;; Todo: Think of a way to start repeated download for every catalog efficiently
(defn set-interval
  "Calls function every ms"
  ^{:deprecated true}
  [callback ms]
  (future (while true (do (try
                            (callback)
                            (println "Recached")
                            (catch Exception e
                              (binding [*out* *err*]
                                (println "Error while updating cache: " e ", retrying in " (/ ms 1000 60) " minutes"))))
                          (Thread/sleep ms)))))

(defn load-config
  "Attempts to load config from file [f].
   Returns loaded config map or nil if failed"
  [f]
  (let [fl (io/as-file f)]
    (when (.exists fl)
      (with-open [r (io/reader fl)]
        (edn/read (java.io.PushbackReader. r))))))

(defn config-fill-board-defaults
  "Fills every enabled board with default config values"
  [config]
  (let [defaults (:boards-defaults config)]
    (dissoc (update-in config
                       '(:boards-enabled)
                       (fn [mp]
                         (u/fmap (fn [k v]
                                   (u/map-apply-defaults v defaults))
                                 mp)))
            :boards-defaults)))

(defn get-some-config
  "Attempts to get config somehow,
   first from [custom-file], if it's nil,
   then from ./config.edn file.
   If is neither exists, default internal one is used."
  [custom-file]
  (config-fill-board-defaults
   ;; TODO: There has to be try/catch for when file is invalid edn
   ;; This is gonna be done when config validation comes in Beta 2
   (let [file-to-try (u/nil?-else custom-file
                                  "./config.edn")]
     (u/when-else (load-config file-to-try)
      CONFIG-DEFAULT))))

(defn -main
  "Entry point, starts webserver"
  [& args]
  (let [parsed-args (parse-opts args cli-options)
        options (get parsed-args :options)]
    (when-let [err (get parsed-args :errors)]
      (println "Error: " err)
      (System/exit 1))
    (when (get options :version)
      (println "RSS Thread Watcher " VERSION " Licensed under AGPL-3.0-only")
      (System/exit 0))
    (when (get options :help)
      (println "RSS Thread Watcher help:\n" (get parsed-args :summary))
      (System/exit 0))
    (when (get options :print-default-config)
      (println ";;Default internal config file from RSS Thread Watcher " VERSION)
      (clojure.pprint/pprint CONFIG-DEFAULT)
      ;; In case someone was copying by hand, this might be useful
      (println ";;END of Default internal config file")
      (System/exit 0))

    (let [config (get-some-config (:config options))]
      ;; TODO: probably refactor to use separate config.clj file when validation will be added
      ;; Init the few globals we have
      (reset! watcher/GLOBAL-CONFIG config)
      (reset! feed/boards-enabled-cache (set (keys (get config :boards-enabled))))
      (reset! watcher/chod-threads-cache (watcher/generate-chod-cache-structure config))
      (clojure.pprint/pprint config)
      (jetty/run-jetty (rp/wrap-params feed/http-handler) {:port (:port CONFIG-DEFAULT)
                                                           :join? true}))))

;; Docs: https://github.com/ring-clojure/ring/wiki/Getting-Started
(defn repl-main
  "Development entry point"
  []
  (jetty/run-jetty (rp/wrap-params #'feed/http-handler)
                   {:port (:port CONFIG-DEFAULT)
                    ;; Dont block REPL thread
                    :join? false}))
;; (repl-main)
;; Single cache update for repl
;; (watcher/update-thread-cache! (:target CONFIG) (:starting-page CONFIG))
