;; Copyright (C) 2024  Felisp
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Affero General Public License as published by
;; the Free Software Foundation, version 3 of the License.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Affero General Public License for more details.
;;
;; You should have received a copy of the GNU Affero General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

(ns rss-thread-watch.watcher
  "Functions for monitoring catalog and maintaining list of near-death threads"
  (:require [clojure.java.io :as io]
            [clojure.data.json :as js])
  (:gen-class))

(def GLOBAL-CONFIG
  "Global config with defaults for missing entires"
  ;; I know globals are ew in Clojure but I don't know any
  ;; better way of doing this
  (atom nil))

(def chod-threads-cache
  "Cached map of threads that have CHanceOfDeath > configured"
  (atom {}))

(defn generate-chod-cache-structure
  "Generates initial structure for global cache
   Structure is returned, you have to set it yourself"
  [config]
  (let [ks (keys (:boards-enabled config))]
    (zipmap ks
            (repeatedly (count ks) #(atom nil)))))

(defn process-page
  "Procesess every thread in page, leaving only relevant information
   (title no chod)"
  ([threads-to-index threads-total starting-index] (process-page threads-to-index threads-total starting-index (transient [])))
  ([remaining-threads threads-total index ret]
   (if (empty? remaining-threads)
     (persistent! ret)
     (let [thread (first remaining-threads)]
       (recur (rest remaining-threads)
              threads-total
              (inc index)
              (conj! ret {:title (or (:sub thread) ;We use thread title if thread has it
                                     (:com thread) ;we use body if thread has it
                                     "")           ;Thread has neither, this prevents null pointer
                          :no (:no thread)
                          :chod (* 100 (float (/ index threads-total)))}))))))

(defn build-cache
  "Build cache of near-death threads so the values don't have to be recalculated on each request."
  [pages-to-index pages-total threads-per-page threads-total]
  {:time (System/currentTimeMillis)
   :data (vec (flatten (map (fn [single-page]
                              ;; We have to (dec page-number) bcs otherwise we would get the total number of threads
                              ;; including the whole page of threads
                              (let [page-number (dec (:page single-page))] ;        inc to get to the actuall page
                                (process-page (:threads single-page) threads-total (inc (* page-number threads-per-page)))))
                            pages-to-index)))})

(defn update-board-cache!
  "Updates cache of near-death threads. Writes to chod-threads-cache as side effect.
   [url] - Url to download data from
   [board] - Board to assign cached data to, it's existence is NOT checked here
   [starting-page] - From which page consider threads to be fit for near-death cache
   THIS FUNCTION WRITES TO chod-threads-cache
   Returns :data part of [board] cache"
  [url board starting-page]
  ;; Todo: surround with try so we can timeout, 40x and other stuff
  (let [catalog (with-open [readr (io/reader url)]
                  (js/read readr :key-fn keyword))
        pages-total (count catalog)
        ;; universal calculation for total number of threads:
        ;; (pages-total -1) * threadsPerPage + threadsOnLastpage ;;accounts for boards which have stickied threads making them have 11pages
        threads-per-page (count (:threads (first catalog))) ;;                    TODO: last could be remade to peek if it's a vector
        threads-total (+ (* threads-per-page (dec pages-total)) (count (:threads (last catalog)))) ;; Todo: Yeah, maybe this calculation could be refactored into let
        to-index (filter (fn [item]
                           (<= starting-page (:page item))) catalog)]
    ;; TODO: there absolutely must be try catch for missing - not enabled boards,
    ;; This is probably resolved now, but keeping it just in case
    ;; This will return nill and that fuck everything up
    (println "Refreshed cache for " board)
    (reset! (get @chod-threads-cache board)
            (build-cache to-index pages-total threads-per-page threads-total))))

(defn board-enabled?
  "Checks whether board is enabled in config"
  [board config]
  (contains? board (keys (get config :boards-enabled))))

(defn get-board-url
  "Gets board url from :target if "
  [board config]
  ;; TODO: jesus, this needs sanitization and should be probably crafted by some URL class
  (str (get-in config [:boards-enabled board :target]) board "catalog.json"))

(defn get-thread-data
  "Gets thread cache for given board.
   If board is lazy loaded, downloads new one if needed.

   MAY CAUSE WRITE TO chod-thread-cache IF NECCESARRY"
  [board config]
  (let [refresh-rate (* 1000 (get-in config `(:boards-enabled ~board :refresh-rate)))
        {data :data
         time-downloaded :time
         :or {time-downloaded 0}
         :as board-atom } @(get @chod-threads-cache board)
        ;; TODO: This also makes it implictly lazy-load -> if disabled make the check here
        time-to-update? (or (nil? board-atom)
                            (> (System/currentTimeMillis) (+ refresh-rate time-downloaded)))]
    (if time-to-update?
      (update-board-cache! (get-board-url board config) board (get-in config [:boards-enabled board :starting-page]))
      @(get @chod-threads-cache board))))
